#!/usr/bin/env python3
#
# Copyright 2019. Clumio, Inc.

import boto3
import click
import pandas as pd
from botocore.exceptions import ClientError
import openpyxl
import datetime
import pprint as pp


@click.group()
def cli():
    '''

    AWS backup sizing tool by Clumio, inc.

    Maintained by Luke McCleary | Luke@clumio.com

    Generates a report using the AWS Python SDK to import stats from
    EC2, EBS, RDS etc. 

    Example 1: Let the script prompt for required inputs: \n
        Python3 -m calipers report

    Example 2: Specify output file name and a geo: \n
        Python3 -m calipers report -o filename.xlsx -g eu

    Example 3: Specify output file name and multiple geos: \n
        Python3 -m calipers report -o filename.xlsx -g eu -g am

    '''
    pass


@cli.command()
@click.option('-id', '--key-id', prompt=True)
@click.option('-sk', '--secret-key', prompt=True, hide_input=True)
# @click.option('-sr', '--single-report', default=None)
@click.option('-o', '--outfile', default='Clumio_Calipers.xlsx')
@click.option('-g', '--geo', multiple=True, required=True, default=['am'],
              type=click.Choice(['eu', 'am', 'apac'], case_sensitive=True))           
def report(key_id, secret_key, geo, outfile):
    '''
    Generate an excel report. Defaults to North and South American
    regions, non-govcloud. Other regions, or for multiple regions,
    us the --geo or -g flags.

    Americas = am \n
    Europe = eu \n
    Asia = apac \n

    '''
    americas = ['us-east-1', 'us-east-2', 'us-west-1',
                'us-west-2', 'ca-central-1', 'sa-east-1']
    europe = ['eu-central-1', 'eu-west-1', 'me-south-1',
              'eu-west-2', 'eu-west-3', 'eu-north-1']
    asiapac = ['ap-east-1', 'ap-south-1',
               'ap-northeast-3', 'ap-northeast-2', 'ap-southeast-1',
               'ap-southeast-2', 'ap-northeast-1']

    regions = []
    if 'am' in geo:
        for i in americas:
            regions.append(i)
    if 'eu' in geo:
        for i in europe:
            regions.append(i)
    if 'apac' in geo:
        for i in asiapac:
            regions.append(i)

    print('\nReporting on the following AWS Regions' + str(regions) + '\n')

    '''initialize dictionaries variables'''
    ec2_dict, ebs_vols_dict, ebs_snap_dict = [], [], []

    '''Regional iterative with connection to ec2'''
    for r in regions:
        try:
            print('\nAttmpting connection to ' + str(r))
            ec2_conn = aws_connect(key_id, secret_key, 'ec2', r)
            print('\nSuccessful connection to EC2 in ' + str(r))
        except ClientError as e:
            print('\nFailed to connect to EC2 in ' + str(r) + '\n' + str(e))
            continue

        try:
            ec2 = ec2_report(ec2_conn, r, ec2_dict)
            ebs, vols_in_region = ebs_report(ec2_conn, r, ebs_vols_dict)
            ebs_snaps = ebs_snap_report(ec2_conn,
                                        vols_in_region, r, ebs_snap_dict)
        except Exception as e:
            print('A general error occured returning data from AWS\n' + str(e))

    '''convert returned dicts to dataframes'''
    ec2_df, ebs_df, ebs_snap_df = create_dataframes(ec2, ebs, ebs_snaps)

    '''pandas commands to merge reports into excel workbooks by tabs'''
    try:
        writer = pd.ExcelWriter(outfile)
        ec2_df.to_excel(writer, 'Running_Stopped_EC2')
        ebs_df.to_excel(writer, 'EBS_Volumes')
        ebs_snap_df.to_excel(writer, 'EBS_Snaphots')
        writer.save()
    except Exception as e:
        print('\nError Writing to Excel File\n' + str(e))


def ec2_report(ec2_conn, region, ec2_dict):
    '''initialize variables'''
    response = []
    ec2_instances = []

    pag = ec2_conn.get_paginator('describe_instances')
    for page in pag.paginate(Filters=[
            {
                'Name': 'instance-state-name',
                'Values': ['running', 'stopped']
            },

    ], PaginationConfig={
            'PageSize': 150
    }
    ):
        for i in page['Reservations']:
            response.append(i)

    for d in range(len(response)):
        for i in range(len(response[d]['Instances'])):
            ec2_instances.append(response[d]['Instances'][i])
    print('\nFound ' + str(len(ec2_instances)) + ' Running or Stopped EC2 '
          'Instances in ' + str(region))

    try:
        '''iterate through regions appending data to datastructure'''
        for i in range(len(ec2_instances)):
            vols = []

            '''store volume IDs in local variable for loop'''
            for v in range(len(ec2_instances[i]['BlockDeviceMappings'])):
                vols.append(ec2_instances[i]['BlockDeviceMappings'][v]['Ebs']
                            ['VolumeId'])

            '''write ec2 data to data structure'''
            ec2_dict.append({
                'InstanceId': ec2_instances[i]['InstanceId'],
                'Type': ec2_instances[i]['InstanceType'],
                'Region': str(region),
                'AZ': ec2_instances[i]['Placement']['AvailabilityZone'],
                'Power-State': ec2_instances[i]['State']['Name'],
                'Attached-Volumes': str(vols)
            })
    except Exception as e:
        print('\nNo EC2 Instances found in ' + str(region) + '\n' + str(e))

    return ec2_dict


def ebs_report(ec2_conn, region, ebs_vols_dict):
    response = []
    ebs_vols = []
    vols_in_region = []

    '''create paginator and build ebs_vols reponse dict'''
    pag = ec2_conn.get_paginator('describe_volumes')

    for page in pag.paginate(Filters=[
            {
                'Name': 'status',
                'Values': ['available', 'in-use']
            }
        ],
        PaginationConfig={
            'PageSize': 150
    }
    ):
        for i in page['Volumes']:
            response.append(i)
    for d in range(len(response)):
        ebs_vols.append(response[d])

    print('\nFound ' + str(len(ebs_vols)) + ' Available or In-Use '
          'EBS volumes in ' + str(region))

    '''Store list of active volumes for subsequent snapshot query'''
    for i in range(len(ebs_vols)):
        vols_in_region.append(ebs_vols[i]['VolumeId'])

    '''write ebs data to data structure'''
    for i in range(len(ebs_vols)):
        ebs_vols_dict.append({
            'VolumeId': ebs_vols[i]['VolumeId'],
            'Size (GB)': ebs_vols[i]['Size'],
            'Region': str(region),
            'AZ': ebs_vols[i]['AvailabilityZone'],
            'State': ebs_vols[i]['State'],
            'Volume Snapshots': ebs_vols[i]['SnapshotId']
        })

    return ebs_vols_dict, vols_in_region


def ebs_snap_report(ec2_conn, vol_list, region, ebs_snap_dict):
    '''get snaps associated with vols_in_region'''
    try:
        for v in vol_list:
            response = ec2_conn.describe_snapshots(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': [v],
                }
            ])['Snapshots']

            '''write ebs_snap data to data structure'''
            for i in range(len(response)):
                awstime = response[i]['StartTime']
                snaptime = awstime.strftime("%d-%b-%Y (%H:%M:%S.%f)")
                ebs_snap_dict.append({
                    'SnapshotId': response[i]['SnapshotId'],
                    'Source Vol ID': response[i]['VolumeId'],
                    'Source Vol Size (GB)': response[i]['VolumeSize'],
                    'Region': str(region),
                    'Snap DateTime': snaptime
                })

    except Exception as e:
        print('\nError locating EBS Snaps from' + str(region) + str(e))

    return ebs_snap_dict


def create_dataframes(ec2_dict, ebs_vols_dict, ebs_snap_dict):
    '''create ec2 dataframe'''
    try:
        ec2_df = pd.DataFrame(ec2_dict)
        ec2_df.set_index('InstanceId', inplace=True)
    except Exception as e:
        print('\nNo EC2 instances were located in the regions specified' + str(e))
        ec2_df = pd.DataFrame(columns=['InstanceId'])

    '''create ebs volme dataframe'''
    try:
        ebs_df = pd.DataFrame(ebs_vols_dict)
        ebs_df.set_index('VolumeId', inplace=True)
    except Exception:
        print('No EBS volumes were located in the regions specified')
        ebs_df = pd.DataFrame(columns=['Volumeid'])

    '''create ebs_snap dataframe'''
    try:
        ebs_snap_df = pd.DataFrame(ebs_snap_dict)
        ebs_snap_df.set_index('SnapshotId', inplace=True)
    except Exception as e:
        print('\nNo EBS Snapshots were located in the regions specified\n' + str(e))
        ebs_snap_df = pd.DataFrame(columns=['SnapshotId'])
    return ec2_df, ebs_df, ebs_snap_df


def rds_report(key_id, secret_key, region):
    for r in region:
        rds_conn = aws_connect(key_id, secret_key, 'ec2', region)
    return


def s3_report(key_id, secret_key, region):
    for r in region:
        s3_conn = aws_connect(key_id, secret_key, 's3', region)
    return


def efs_report(key_id, secret_key, region):
    for r in region:
        efs_conn = aws_connect(key_id, secret_key, 'efs', region)
    return


def aws_connect(key_id, secret_key, service, region):
    client = boto3.client(
        service,
        aws_access_key_id=key_id,
        aws_secret_access_key=secret_key,
        region_name=region
    )

    return client


def get_bucket_list(region, s3_conn):
    s3_data = s3_conn.list_buckets()
    s3_bucket_list = []
    for i in range(len(s3_data['Buckets'])):
        s3_bucket_list.append(s3_data['Buckets'][i]['Name'])

    return s3_bucket_list


def get_bucket_bytes(bucket_name, cw_conn):
    response = cw_conn.get_metric_statistics(
        Namespace="AWS/S3",
        MetricName="BucketSizeBytes",
        Dimensions=[
            {
                "Name": "BucketName",
                "Value": bucket_name
            },
            {
                "Name": "StorageType",
                "Value": "StandardStorage"
            }
        ],
        StartTime=datetime.now() - timedelta(days=1),
        EndTime=datetime.now(),
        Period=86400,
        Statistics=['Average']
    )
    return response


if __name__ == '__main__':
    cli()
